#!/usr/bin/python3

from random import shuffle

from player import Player
import ai


def spies_for_players(player_count):
    return {
        5: 2,
        6: 2,
        7: 3,
        8: 3,
        9: 3,
        10: 4
    }[player_count]


def mission_size(player_count, mission_num):
    return {
        5: [2, 3, 2, 3, 3],
        6: [2, 3, 4, 3, 4],
        7: [2, 3, 3, 4, 4],
        8: [3, 4, 4, 5, 5],
        9: [3, 4, 4, 5, 5],
        10: [3, 4, 4, 5, 5]
    }[player_count][mission_num - 1]


def required_failures(player_count, mission_num):
    if mission_num == 4 and player_count >= 7:
        return 2
    else:
        return 1


class Shared(object):

    def __init__(self):
        pass


class AdvanceMission(object):

    def __init__(self, shared):
        self.shared = shared


class AssembleTeam(object):

    def __init__(self, shared, team_size):
        self.shared = shared
        self.team_size = team_size


class PerformMission(object):

    def __init__(self, shared):
        self.shared = shared


def main(player_count):
    spy_count = spies_for_players(player_count)
    resistance_count = player_count - spy_count
    print("Will have {} spies and {} resistance".format(spy_count, resistance_count))
    role_deck = ["RES"] * resistance_count + ["SPY"] * spy_count
    shuffle(role_deck)
    print(role_deck)
    players = [Player(player_count, i, r) if r == "SPY"
               else ai.Player(player_count, i, r)
               for i, r in enumerate(role_deck)]

    spy_nums = [i for i, r in enumerate(role_deck) if r == "SPY"]

    for i, r in enumerate(role_deck):
        if r == "SPY":
            players[i].reveal_spies(spy_nums)

    shared = Shared()

    wins = 0
    losses = 0
    captain = 0
    team_failures = 0
    mission = 1
    team = None
    approved = None

    def advance_mission():
        team_size = mission_size(player_count, mission)
        print("Mission {} needs {} members".format(mission, team_size))
        return AssembleTeam(shared, team_size)

    def assemble_team():
        nonlocal approved, captain, team, team_failures

        print("Captain is {}".format(captain))
        team = players[captain].propose_team(state.team_size)
        print("He proposed {}".format(team))

        votes = [p.approve_team(team) for p in players]
        approvals = sum(votes)
        approved = approvals > state.team_size / 2

        print("Team {}: {} for, {} against".format(
            {True: "approved", False: "rejected"}[approved],
            approvals, player_count - approvals
        ))

        for p in players:
            p.observe_team_vote(approved, votes)

        captain = (captain + 1) % player_count

        if approved:
            team_failures = 0
            return PerformMission(shared)
        else:
            team_failures += 1
            return AssembleTeam(shared, state.team_size)

    def perform_mission():
        nonlocal wins, losses, mission

        failures = 0
        for i in team:
            if role_deck[i] == "SPY":
                if not players[i].perform_mission():
                    failures = failures + 1

        success = failures < required_failures(player_count, mission)

        for p in players:
            p.observe_mission(success, failures)

        if success:
            print("Mission succeeded")
            wins += 1
        else:
            print("Mission had {} failures".format(failures))
            losses += 1

        mission += 1
        return AdvanceMission(shared)

    state = AdvanceMission(shared)

    while True:
        if isinstance(state, AdvanceMission):
            state = advance_mission()
        elif isinstance(state, AssembleTeam):
            state = assemble_team()
        elif isinstance(state, PerformMission):
            state = perform_mission()

        if team_failures == 5:
            print("The spies won!")
            for p in players:
                p.observe_game(False)
            return
        elif wins >= 3:
            print("The Resistance won!")
            for p in players:
                p.observe_game(True)
            return
        elif losses >= 3:
            print("The Spies won!")
            for p in players:
                p.observe_game(False)
            return


if __name__ == '__main__':
    main(10)
